/*
** Module   : FieldReference.java
** Abstract : Buffer field reference which allows deferred read/write access
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ------------------Description-------------------
** 001 ECF 20051112   @23388 Created initial version. Resolves a buffer
**                           field reference value at runtime. Used as a
**                           query substitution parameter for field
**                           references which can only be resolved as the
**                           query is executing and has partial results.
** 002 ECF 20051216   @23756 Added getValue method. Returns resolved value
**                           but reports a warning to the UI (via the
**                           error manager) if resolution fails.
** 003 ECF 20060104   @23822 Implement Resolvable interface. Allows this
**                           class to be used generically with util APIs.
** 004 ECF 20060202   @24223 Implement Accessor interface. Field access is
**                           now read/write instead of read-only.
** 005 ECF 20060307   @25015 Added new constructor and get method variant.
**                           These permit late binding of a DMO instance.
** 006 ECF 20060323   @25226 Added support for case-insensitive field
**                           references used as query substitution
**                           parameters. New constructor variants accept
**                           an uppercase flag.
** 007 GES 20060907   @29285 Setting an extent field was broken. The index
**                           had to be the first parameter instead of the
**                           second.
** 008 ECF 20060927   @29970 Added special exception handling in set() and
**                           getObject(). ConditionException wrapped in an
**                           InvocationTargetExcpetion is unwrapped and
**                           rethrown.
** 009 ECF 20061027   @30749 Replaced Object with DataModelObject for DMO
**                           arguments to methods/constructors. Required
**                           for compile-time type safety.
** 010 ECF 20070416   @33022 Integrated user interrupt handling.
** 011 ECF 20080528   @38480 Integrated generics.
** 012 ECF 20080606   @38645 Fixed FieldReference(Class<?>, String) c'tor
**                           variant. It was not properly setting the
**                           uppercase instance variable in the case of
**                           references to case-insensitive properties.
**                           Also added support for the reserved 'id'
**                           property in getGetter() method.
** 013 CA  20080815   @39458 Support API change in DBUtils.
** 014 ECF 20090701   @43217 Enabled lazy initialization of backing record
**                           buffer. Database instance variable is extracted
**                           from buffer lazily, rather than at construction.
** 015 SVL 20090723   @43342 Added getPropertyName() function.
** 016 ECF 20090819   @43695 Downgraded some log messages from ERROR to DEBUG.
** 017 GES 20090828   @43799 Don't unconditionally shift into warning mode as
**                           that would disable warning mode on exit from the
**                           getter even if it had already been enabled.
** 018 AIL 20121206          Update from character to TextOps.
** 019 VMN 20130207          Extended from FieldInterface and added support for attributes
**                           BUFFER-VALUE (KW_BUF_VAL)
**                           COLUMN-LABEL (KW_COL_LAB)
**                           EXTENT (KW_EXTENT)
**                           FORMAT (KW_FORMAT)
**                           STRING-VALUE (KW_STR_VAL)
*/

package com.goldencode.p2j.persist;

import java.lang.reflect.*;
import java.util.*;
import org.apache.commons.logging.*;
import com.goldencode.p2j.*;
import com.goldencode.p2j.util.*;

/**
 * A readable, writable reference to a buffer field, used in cases where the
 * field needs to be accessed at some arbitrary point in time after it is
 * referenced in application code.  For example, an instance of this class
 * would be used as a substitution parameter in a {@link CompoundQuery}, when
 * an inner loop where clause of a multi-table query refers to a field within
 * a buffer retrieved in an outer loop.
 * <p>
 * This class provides both read and write capability to the backing field,
 * insofar as the DMO has both an accessor (getter) and a mutator (setter)
 * method defined for the specified property.  Both indexed and non-indexed
 * versions of these methods are supported.  It is assumed that the methods
 * have bean-like signatures.  Specifically, the following, public signature
 * patterns are assumed:
 * <p>
 * <dt>Simple Getter:</dt>
 * <dd>
 * <code>&lt;BaseDataType subclass&gt; get&lt;Property&gt;()</code>
 * </dd>
 * <dt>Indexed Getter:</dt>
 * <dd>
 * <code>&lt;BaseDataType subclass&gt; get&lt;Property&gt;(int)</code>
 * </dd>
 * <dt>Simple Setter:</dt>
 * <dd>
 * <code>void set&lt;Property&gt;(&lt;BaseDataType subclass&gt;)</code>
 * </dd>
 * <dt>Indexed Setter:</dt>
 * <dd>
 * <code>void set&lt;Property&gt;(&lt;BaseDataType subclass&gt;, int)</code>
 * </dd>
 * <p>
 */
public final class FieldReference
implements Resolvable,
           Accessor,
           FieldInterface
{
   /** Logger */
   private static final Log LOG = Utils.getLog(FieldReference.class);
   
   /** Maps DMO interfaces to maps of property names to getter methods */
   private static final Map<Class<?>, Map<String, Method>> getterCache =
      new HashMap<Class<?>, Map<String, Method>>();
   
   /** Maps DMO interfaces to maps of property names to setter methods */
   private static final Map<Class<?>, Map<String, Method>> setterCache =
      new HashMap<Class<?>, Map<String, Method>>();
   
   /** Property name */
   private final String property;
   
   /** Method used to resolve this field reference at query execution time */
   private final Method getter;
   
   /** Method used to resolve this field reference at query execution time */
   private final Method setter;
   
   /** Optional index, used with a DMO indexed getter method */
   private final NumberType index;
   
   /** The associated database */
   private Database database = null;

   /** DMO or DMO proxy instance;  the referent of the referencing method */
   private Object dmo = null;
   
   /** Flag to force resolved character result to uppercase/right-trim */
   private boolean uppercase = false;
   
   /**
    * Create a new field reference given a DMO proxy and property name.  A
    * non-indexed getter method will be used to resolve the field.
    *
    * @param   dmo
    *          DMO proxy.
    * @param   property
    *          Target property name.
    */
   public FieldReference(DataModelObject dmo, String property)
   {
      this(dmo, property, null);
   }
   
   /**
    * Create a new field reference given a DMO proxy and property name.  A
    * non-indexed getter method will be used to resolve the field.
    *
    * @param   dmo
    *          DMO proxy.
    * @param   property
    *          Target property name.
    * @param   uppercase
    *          <code>true</code> to force a <code>character</code> value
    *          returned by {@link #getObject} to be uppercased/right-trimmed;
    *          <code>false</code> to leave the result unchanged.
    */
   public FieldReference(DataModelObject dmo,
                         String property,
                         boolean uppercase)
   {
      this(dmo, property, uppercase, null);
   }
   
   /**
    * Create a new field reference given a DMO interface and property name.
    * A non-indexed getter method will be used to resolve the field.  A DMO
    * instance must be provided at resolution time, using the {@link
    * #get(Persistable)} variant of the <code>get</code> method.
    *
    * @param   database
    *          The associated database.
    * @param   dmoIface
    *          Interface of the target DMO.
    * @param   property
    *          Target property name
    */
   FieldReference(Database database, Class<?> dmoIface, String property)
   {
      this.database = database;
      this.property = property;
      this.getter = getGetter(dmoIface, property);
      this.setter = null;
      this.index = null;
      
      if (character.class.equals(getter.getReturnType()))
      {
         String schema = DatabaseManager.getSchemaByInterface(dmoIface);
         if (!DMOIndex.isCaseSensitive(schema, dmoIface, property))
         {
            this.uppercase = true;
         }
      }
   }
   
   /**
    * Create a new field reference given a DMO proxy, a property name, and
    * an index for the property.  An indexed getter method will be used to
    * resolve the field.
    *
    * @param   dmo
    *          DMO proxy.
    * @param   property
    *          Target property name.
    * @param   index
    *          Index to pass to property's getter method at resolve time.
    */
   public FieldReference(DataModelObject dmo, String property, int index)
   {
      this(dmo, property, false, new integer(index));
   }
   
   /**
    * Create a new field reference given a DMO proxy, a property name, and
    * an index for the property.  An indexed getter method will be used to
    * resolve the field.
    *
    * @param   dmo
    *          DMO proxy.
    * @param   property
    *          Target property name.
    * @param   uppercase
    *          <code>true</code> to force a <code>character</code> value
    *          returned by {@link #getObject} to be uppercased/right-trimmed;
    *          <code>false</code> to leave the result unchanged.
    * @param   index
    *          Index to pass to property's getter method at resolve time.
    */
   public FieldReference(DataModelObject dmo,
                         String property,
                         boolean uppercase,
                         int index)
   {
      this(dmo, property, uppercase, new integer(index));
   }
   
   /**
    * Create a new field reference given a DMO proxy, a property name, and
    * an index for the property.  An indexed getter method will be used to
    * resolve the field.
    *
    * @param   dmo
    *          DMO proxy.
    * @param   property
    *          Target property name.
    * @param   index
    *          Index to pass to property's getter method at resolve time.
    *
    * @throws  IllegalArgumentException
    *          if field reference is not recognized.
    */
   public FieldReference(DataModelObject dmo,
                         String property,
                         NumberType index)
   {
      this(dmo, property, false, index);
   }
   
   /**
    * Create a new field reference given a DMO proxy, a property name, and
    * an index for the property.  An indexed getter method will be used to
    * resolve the field.
    *
    * @param   dmo
    *          DMO proxy.
    * @param   property
    *          Target property name.
    * @param   uppercase
    *          <code>true</code> to force a <code>character</code> value
    *          returned by {@link #getObject} to be uppercased/right-trimmed;
    *          <code>false</code> to leave the result unchanged.
    * @param   index
    *          Index to pass to property's getter method at resolve time.
    *
    * @throws  IllegalArgumentException
    *          if field reference is not recognized.
    */
   public FieldReference(DataModelObject dmo,
                         String property,
                         boolean uppercase,
                         NumberType index)
   {
      this.dmo = dmo;
      this.property = property;
      RecordBuffer buffer = RecordBuffer.get(dmo);
      Class<?> dmoIface = buffer.getDMOInterface();
      this.getter = getGetter(dmoIface, property);
      this.setter = getSetter(dmoIface, property);
      this.index = index;
      this.uppercase = uppercase;
      
      if (getter == null || setter == null)
      {
         throw new IllegalArgumentException(
            "Invalid field reference [" + buffer.getDMOName() + "." +
            property + (index == null ? "" : ":" + index.toStringMessage()) +
            "]");
      }
      
      if (uppercase && !getter.getReturnType().equals(character.class))
      {
         throw new IllegalArgumentException(
            "Only a character-based FieldReference can be set to uppercase");
      }
   }
   
   /**
    * Convenience method to produce a <code>BaseDataType</code> instance
    * initialized to the unknown value.  An attempt is made to return an
    * object of the correct type (i.e., the getter method's return type).
    * If this fails for some reason, a new instance of <code>unknown</code>
    * is returned.
    *
    * @param   getter
    *          Getter method whose return type is used to determine the type
    *          of object to return.
    *
    * @return  An object whose type matches the return type of the getter
    *          method, or an instance of <code>unknown</code>.
    */
   static BaseDataType unknownValue(Method getter)
   {
      BaseDataType value;
      try
      {
         value = (BaseDataType) getter.getReturnType().newInstance();
         value.setUnknown();
      }
      catch (InstantiationException exc)
      {
         value = new unknown();
      }
      catch (IllegalAccessException exc)
      {
         throw new RuntimeException(exc);
      }
      
      return value;
   }
   
   /**
    * Look up the getter method for the specified property and interface.
    * Introspection is used the first time this method is invoked for a
    * new DMO type;  thereafter, the getter is retrieved from a cache.
    * <p>
    * For the reserved <code>id</code> property, {@link Persistable} interface
    * is substituted for <code>dmoIface</code>.
    *
    * @param   dmoIface
    *          Referent DMO interface.
    * @param   property
    *          Name of target property (i.e., field).
    *
    * @return  Getter method, or <code>null</code> if <code>property</code>
    *          does not match any property name in the DMO API.
    */
   private static Method getGetter(Class<?> dmoIface, String property)
   {
      Map<String, Method> propMap = null;
      Class<?> iface = ("id".equals(property) ? Persistable.class : dmoIface);
      
      synchronized (getterCache)
      {
         propMap = getterCache.get(iface);
         if (propMap == null)
         {
            propMap = PropertyHelper.gettersByProperty(iface);
            getterCache.put(iface, propMap);
         }
      }
      
      return propMap.get(property);
   }
   
   /**
    * Look up the setter method for the specified property and interface.
    * Introspection is used the first time this method is invoked for a
    * new DMO type;  thereafter, the setter is retrieved from a cache.
    *
    * @param   dmoIface
    *          Referent DMO interface.
    * @param   property
    *          Name of target property (i.e., field).
    *
    * @return  Setter method, or <code>null</code> if <code>property</code>
    *          does not match any property name in the DMO API.
    */
   private static Method getSetter(Class<?> dmoIface, String property)
   {
      Map<String, Method> propMap = null;
      synchronized (setterCache)
      {
         propMap = setterCache.get(dmoIface);
         if (propMap == null)
         {
            propMap = PropertyHelper.settersByProperty(dmoIface);
            setterCache.put(dmoIface, propMap);
         }
      }
      
      return propMap.get(property);
   }
   
   /**
    * Determine the data type that will be returned when this field reference
    * is resolved.
    *
    * @return  Class of the appropriate data type.
    */
   public Class<?> getType()
   {
      return getter.getReturnType();
   }
   
   /**
    * Resolve the field reference's value, bracketed by calls to the error
    * manager to first enable, then disable, warning mode.  This is meant to
    * support the need to continue processing after resolution of a field
    * reference fails, due to an empty record buffer.
    *
    * @return  Current value of the target property in the referent, or the
    *          unknown value if this resolution fails.
    */
   public BaseDataType getValue()
   {
      boolean warning = ErrorManager.isWarningMode();
      
      if (!warning)
      {
         ErrorManager.warningModeEnable();
      }
      
      try
      {
         return get();
      }
      
      finally
      {
         if (!warning)
         {
            ErrorManager.warningModeDisable();
         }
      }
   }
   
   /**
    * Resolve the current value of the referent field by invoking the getter
    * method for the target property.  Note that this may return the unknown
    * value.
    *
    * @return  Current value of the target property in the referent.
    */
   public BaseDataType resolve()
   {
      return get();
   }
   
   /**
    * Resolve the current value of the referent field by invoking the getter
    * method for the target property.  Note that this may return the unknown
    * value.
    *
    * @return  Current value of the target property in the referent.
    *
    * @throws  ErrorConditionException
    *          if any persistence level error occurs getting the value from
    *          the DMO.
    * @throws  IllegalStateException
    *          if the backing property is not read accessible (i.e., no getter
    *          method was found).
    * @throws  RuntimeException
    *          if the backing, getter method cannot be accessed or if there
    *          is some other error at the invocation target.
    */
   public BaseDataType get()
   {
      return (BaseDataType) getObject();
   }
   
   /**
    * Set the current value of the referent field by invoking the setter
    * method for the target property.
    *
    * @param   value
    *          New value to set into the field represented by this object.
    *
    * @throws  ErrorConditionException
    *          if this field reference represents an element in an array, but
    *          the index provided at construction is the unknown value;
    *          if any persistence-level error occurs setting the value into
    *          the DMO, including validation errors.
    * @throws  IllegalStateException
    *          if the backing property is not write accessible (i.e., no
    *          setter method was found).
    * @throws  RuntimeException
    *          if the backing, getter method cannot be accessed or if there
    *          is some other error at the invocation target.
    */
   public void set(BaseDataType value)
   {
      if (setter == null)
      {
         throw new IllegalStateException(
            property + " property is not write-accessible [" +
            dmo.getClass().getName() + "]");
      }
      
      Object[] args;
      
      // Handle extent fields.
      if (index != null)
      {
         // An unknown value index constitutes an error.
         if (index.isUnknown())
         {
            ErrorManager.recordOrThrowError(
               26,
               "Array subscript 0 is out of range");
            
            return;
         }
         
         // Set up arguments, including index.
         args = new Object[] { new Integer(index.intValue()), value };
      }
      else
      {
         args = new Object[] { value };
      }
      
      // Invoke the setter method on the proxy.
      try
      {
         setter.invoke(dmo, args);
      }
      catch (InvocationTargetException exc)
      {
         handleException(exc);
         
         Throwable cause = exc.getCause();
         
         if (LOG.isDebugEnabled())
         {
            LOG.debug("Error setting value into field", cause);
         }
         
         if (cause instanceof NumberedException)
         {
            ErrorManager.recordOrThrowError((NumberedException) cause);
         }
         else if (cause instanceof ConditionException)
         {
            throw (ConditionException) cause;
         }
         else
         {
            // Represents programming error.
            throw new RuntimeException(cause);
         }
      }
      catch (IllegalAccessException exc)
      {
         // Represents programming or JVM configuration error.
         throw new RuntimeException(exc);
      }
   }
   
   /**
    * Resolve the current value of the referent field by invoking the getter
    * method for the target property, using the given DMO as the target.
    * Note that this may return <code>null</code> (or the unknown value for
    * <code>BaseDataTypes</code> return types).
    *
    * @param   dmo
    *          Target object which is used as the method's referent.
    *
    * @return  Current value of the target property in the referent or
    *          <code>null</code> if no value exists.
    *
    * @throws  ErrorConditionException
    *          if any persistence level error occurs getting the value from
    *          the DMO.
    * @throws  IllegalStateException
    *          if the backing property is not read accessible (i.e., no getter
    *          method was found).
    * @throws  RuntimeException
    *          if the backing, getter method cannot be accessed or if there
    *          is some other error at the invocation target.
    */
   Object get(DataModelObject dmo)
   {
      this.dmo = dmo;
      
      return getObject();
   }
   
   /**
    * Resolve the current value of the referent field by invoking the getter
    * method for the target property, using the given DMO as the target.
    * Note that this may return <code>null</code> (or the unknown value for
    * <code>BaseDataTypes</code> return types).
    *
    * @param   dmo
    *          Target object which is used as the method's referent.
    *
    * @return  Current value of the target property in the referent or
    *          <code>null</code> if no value exists.
    *
    * @throws  ErrorConditionException
    *          if any persistence level error occurs getting the value from
    *          the DMO.
    * @throws  IllegalStateException
    *          if the backing property is not read accessible (i.e., no getter
    *          method was found).
    * @throws  RuntimeException
    *          if the backing, getter method cannot be accessed or if there
    *          is some other error at the invocation target.
    */
   Object get(Persistable dmo)
   {
      this.dmo = dmo;
      
      return getObject();
   }

   /**
    * Returns target property name.
    *
    * @return See above.
    */
   String getPropertyName()
   {
      return property;
   }

   /**
    * Resolve the current value of the referent field by invoking the getter
    * method for the target property. Note that this may return
    * <code>null</code> (or the unknown value for <code>BaseDataTypes</code>
    * return types).
    *
    * @return  Current value of the target property in the referent or
    *          <code>null</code> if no value exists.
    *
    * @throws  ErrorConditionException
    *          if any persistence level error occurs getting the value from
    *          the DMO.
    * @throws  IllegalStateException
    *          if the backing property is not read accessible (i.e., no getter
    *          method was found).
    * @throws  RuntimeException
    *          if the backing, getter method cannot be accessed or if there
    *          is some other error at the invocation target.
    */
   private Object getObject()
   {
      if (getter == null)
      {
         throw new IllegalStateException(
            property + " property is not read-accessible [" +
            dmo.getClass().getName() + "]");
      }
      
      Object[] args;
      
      // Handle extent fields.
      if (index != null)
      {
         // An unknown value index propagates to an unknown value result.
         if (index.isUnknown())
         {
            return unknownValue(getter);
         }
         
         // Set index argument.
         args = new Object[] { new Integer(index.intValue()) };
      }
      else
      {
         args = null;
      }
      
      // Invoke the getter method on the proxy and return the result.
      try
      {
         Object result = getter.invoke(dmo, args);
         
         // If marked for uppercase, then we must uppercase and right trim
         // the result.  This is a special case where the result will be used
         // in Progress-compatible comparisons.
         if (uppercase)
         {
            character c = (character) result;
            result = TextOps.toUpperCase(TextOps.rightTrim(c));
         }
         
         return result;
      }
      catch (InvocationTargetException exc)
      {
         handleException(exc);
         
         Throwable cause = exc.getCause();
         
         if (LOG.isDebugEnabled())
         {
            LOG.debug("Error getting value from field", cause);
         }
         
         if (cause instanceof NumberedException)
         {
            ErrorManager.recordOrThrowError((NumberedException) cause);
         }
         else if (cause instanceof ConditionException)
         {
            throw (ConditionException) cause;
         }
         else
         {
            // Represents programming error.
            throw new RuntimeException(cause);
         }
         
         return unknownValue(getter);
      }
      catch (IllegalAccessException exc)
      {
         // Represents programming or JVM configuration error.
         throw new RuntimeException(exc);
      }
   }
   
   /**
    * Handle a thread interruption or database connection error specially,
    * otherwise return normally.
    * 
    * @param   exc
    *          Exception to be inspected.
    * 
    * @see     DBUtils#handleException(Database, Exception)
    */
   private void handleException(Exception exc)
   {
      if (database == null && dmo instanceof DataModelObject)
      {
         RecordBuffer buffer = RecordBuffer.get((DataModelObject) dmo);
         database = buffer.getDatabase();
      }
      
      DBUtils.handleException(database, exc);
   }

   /**
    * Conversion of EXTENT attribute (KW_EXTENT).
    *
    * Getter of EXTENT attribute
    *
    */
   public integer extent()
   {
      // TODO:  implement
      return null;
   }

   /**
    * Conversion of FORMAT attribute (KW_FORMAT).
    *
    * Getter of FORMAT attribute.
    *
    * @return  current value of FORMAT attribute.
    */
   public character format()
   {
      // TODO:  implement
      return null;
   }

   /**
    * Conversion of FORMAT attribute (KW_FORMAT).
    *
    * Setter of FORMAT attribute.
    *
    */
   public void changeFormat(character format)
   {
      // TODO:  implement
   }

   /**
    * Conversion of STRING-VALUE attribute (KW_STR_VAL).
    *
    * Getter of STRING-VALUE attribute.
    *
    * @return  current value of STRING-VALUE attribute.
    */
   public character stringValue()
   {
      // TODO:  implement
      return null;
   }

   /**
    * Conversion of STRING-VALUE attribute (KW_STR_VAL).
    *
    * Getter of STRING-VALUE attribute.
    *
    * @param   index
    *          An integer expression representing a subscript, for fields that
    *          have extents.
    *
    * @return  current value of STRING-VALUE attribute.
    */
   public character stringValue(int index)
   {
      // TODO:  implement
      return null;
   }

   /**
    * Conversion of BUFFER-VALUE attribute (KW_BUF_VAL).
    *
    * Getter of current value of a buffer-field object. If you modify the
    * BUFFER-VALUE attribute, Progress sets the buffer-field object to the
    * new value.
    *
    * @param   index
    *          An INTEGER expression representing a subscript, for fields
    *          that have extents.
    *
    * @return  current value of a buffer-field object.
    */
   public Object value(int index)
   {
      // TODO:  implement
      return null;
   }

   /**
    * Conversion of BUFFER-VALUE attribute (KW_BUF_VAL).
    *
    * Setter of current value of a buffer-field object. If you modify the
    * BUFFER-VALUE attribute, Progress sets the buffer-field object to the
    * new value.
    *
    * @param   index
    *          An INTEGER expression representing a subscript, for fields
    *          that have extents.
    * @param   dmo
    *          New value of a buffer-field object.
    *
    */
   public void changeValue(int index, Object dmo)
   {
      // TODO:  implement
   }

   /**
    * Conversion of BUFFER-VALUE attribute (KW_BUF_VAL).
    *
    * Getter of current value of a buffer-field object. If you modify the
    * BUFFER-VALUE attribute, Progress sets the buffer-field object to the
    * new value.
    *
    * @return  current value of a buffer-field object.
    */
   public Object value()
   {
      // TODO:  implement
      return null;
   }

   /**
    * Conversion of BUFFER-VALUE attribute (KW_BUF_VAL).
    *
    * Setter of current value of a buffer-field object. If you modify the
    * BUFFER-VALUE attribute, Progress sets the buffer-field object to the
    * new value.
    *
    * @param   dmo
    *          New value of a buffer-field object.
    *
    */
   public void changeValue(Object dmo)
   {
      // TODO:  implement
   }

   /**
    * Conversion of COLUMN-LABEL attribute (KW_COL_LAB).
    *
    * Getter of COLUMN-LABEL attribute.
    *
    * @return  current value of COLUMN-LABEL attribute.
    */
   public character columnLabel()
   {
      // TODO:  implement
      return null;
   }

   /**
    * Conversion of COLUMN-LABEL attribute (KW_COL_LAB).
    *
    * Getter of COLUMN-LABEL attribute.

    * @param   columnLabel
    *          New value of a attribute.
    */
   public void changeColumnLabel(character columnLabel)
   {
      // TODO:  implement
   }

}
